(module asplevels-ex-test-bugpatterns scheme
  (require redex
           "../asplevels-ex-core.rkt")
  (provide asplevels-ex-test-bugpatterns)
  
  (define (asplevels-ex-test-bugpatterns λlang λreduction-rules)
    (begin
      (display "Running Bug Patterns tests")
      (newline)
      
      ;; Advice = λ(q c x y) where
      ;; q = original function
      ;; c = pointcut result
      ;; x = first argument 
      ;; y = second argument... and so on
      ;; Advice args numbers must be 2 + original function args      
      
      ;; Late Bind Handler test
      ;; Aspect as Signaler
      
      ; Non level-capturing handler doesn't catch
      ; aspect exception at base level
      (test-->>
       ;From
       λreduction-rules
       (term
        (0
         (stack) 
         (asps (;Level
                0 
                ;Pointcut
                (λ r1 -1 (jpt) (list))
                ;Advice
                (λ r2 -1 (q c) (raise #f))))
         (try ((λ () #t)) with (λ (ex) 42))))
       ;To
       (term
        (0
         (stack) 
         (asps (;Level
                0 
                ;Pointcut
                (λ r1 -1 (jpt) (list))
                ;Advice
                (λ r2 -1 (q c) (raise #f))))
         (raise 1 #f))))
      
      ; Level-capturing handler doesn't catch
      ; aspect exception at base level
      (test-->>
       ;From
       λreduction-rules
       (term
        (0
         (stack) 
         (asps (;Level
                0 
                ;Pointcut
                (λ r1 -1 (jpt) (list))
                ;Advice
                (λ r2 -1 (q c) (raise #f))))
         (try ((λ () #t)) with (λs (ex) 42))))
       ;To
       (term
        (0
         (stack) 
         (asps (;Level
                0 
                ;Pointcut
                (λ r1 -1 (jpt) (list))
                ;Advice
                (λ r2 -1 (q c) (raise #f))))
         (raise 1 #f))))
      
      ; Non-level capturing handler doesn't catch
      ; aspect exception using up
      (test-->>
       ;From
       λreduction-rules
       (term
        (0
         (stack) 
         (asps (;Level
                0 
                ;Pointcut
                (λ r1 -1 (jpt) (list))
                ;Advice
                (λ r2 -1 (q c) (raise #f))))
         (try ((λ () #t)) with (up (λ (ex) 42)))))
       ;To
       (term
        (0
         (stack) 
         (asps (;Level
                0 
                ;Pointcut
                (λ r1 -1 (jpt) (list))
                ;Advice
                (λ r2 -1 (q c) (raise #f))))
         (raise 1 #f))))      
      
      ; Level capturing handler catches aspect exception using up
      (test-->>
       ;From
       λreduction-rules
       (term
        (0
         (stack) 
         (asps (;Level
                0 
                ;Pointcut
                (λ r1 -1 (jpt) (list))
                ;Advice
                (λ r2 -1 (q c) (raise #f))))
         (try ((λ () #t)) with (up (λs (ex) 42)))))
       ;To
       (term
        (0
         (stack) 
         (asps (;Level
                0 
                ;Pointcut
                (λ r1 -1 (jpt) (list))
                ;Advice
                (λ r2 -1 (q c) (raise #f))))
         42)))
      
      
      ;; Late Bind Handler test
      ;; Aspect as Handler
      
      ; Non level-capturing aspect handler:
      ; doesn't catch exception raised in base code
      (test-->>
       ;From
       λreduction-rules
       (term
        (0
         (stack) 
         (asps (;Level
                0 
                ;Pointcut
                (λ r1 -1 (jpt) (list))
                ;Advice
                (λ r2 -1 (q c x)
                  (try (q x) with (λ (ex) 42)))))
         (raise #f)))
       ;To
       (term
        (0
         (stack) 
         (asps (;Level
                0 
                ;Pointcut
                (λ r1 -1 (jpt) (list))
                ;Advice
                (λ r2 -1 (q c x)
                  (try (q x) with (λ (ex) 42)))))
         (raise 0 #f))))
      
      ; Level-capturing aspect handler:
      ; doesn't catch exception raised in base code
      (test-->>
       ;From
       λreduction-rules
       (term
        (0
         (stack) 
         (asps (;Level
                0 
                ;Pointcut
                (λ r1 -1 (jpt) (list))
                ;Advice
                (λ r2 -1 (q c x)
                  (try (q x) with (λs (ex) 42)))))
         (raise #f)))
       ;To
       (term
        (0
         (stack) 
         (asps (;Level
                0 
                ;Pointcut
                (λ r1 -1 (jpt) (list))
                ;Advice
                (λ r2 -1 (q c x)
                  (try (q x) with (λs (ex) 42)))))
         (raise 0 #f))))
      
      ; Non level-capturing aspect handler with down operator:
      ; catches exception raised in base code
      (test-->>
       ;From
       λreduction-rules
       (term
        (0
         (stack) 
         (asps (;Level
                0 
                ;Pointcut
                (λ r1 -1 (jpt) (list))
                ;Advice
                (λ r2 -1 (q c x)
                  (try (q x) with (down (λ (ex) 42))))))
         ((λ(x) (raise x)) 1)))
       ;To
       (term
        (0
         (stack) 
         (asps (;Level
                0 
                ;Pointcut
                (λ r1 -1 (jpt) (list))
                ;Advice
                (λ r2 -1 (q c x)
                  (try (q x) with (down (λ (ex) 42))))))
         (raise 0 1)))) ;Proceeds shift down execution of try block
      
      ; Level-capturing aspect handler with down operator:
      ; catches exception raised in base code
      (test-->>
       ;From
       λreduction-rules
       (term
        (0
         (stack) 
         (asps (;Level
                0 
                ;Pointcut
                (λ r1 -1 (jpt) (list))
                ;Advice
                (λ r2 -1 (q c x)
                  (try (q x) with (down (λs (ex) 42))))))
         ((λ(x) (raise x)) 1)))
       ;To
       (term
        (0
         (stack) 
         (asps (;Level
                0 
                ;Pointcut
                (λ r1 -1 (jpt) (list))
                ;Advice
                (λ r2 -1 (q c x)
                  (try (q x) with (down (λs (ex) 42))))))
         42)))
      
      
      
      
      (test-results))))
